/******************************************************************************
 * Icinga 2                                                                   *
 * Copyright (C) 2018 Herrmann & Lenz Solutions GmbH (https://hl-solutions.de)*
 *                                                                            *
 * This program is free software; you can redistribute it and/or              *
 * modify it under the terms of the GNU General Public License                *
 * as published by the Free Software Foundation; either version 2             *
 * of the License, or (at your option) any later version.                     *
 *                                                                            *
 * This program is distributed in the hope that it will be useful,            *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of             *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              *
 * GNU General Public License for more details.                               *
 *                                                                            *
 * You should have received a copy of the GNU General Public License          *
 * along with this program; if not, write to the Free Software Foundation     *
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.             *
 ******************************************************************************/

#include "hlmon/hlmonreceiver.hpp"
#include "hlmon/hlmonreceiver-ti.cpp"

#include <SimpleAmqpClient/SimpleAmqpClient.h>

#include "icinga/externalcommandprocessor.hpp"
#include "base/configtype.hpp"
#include "base/logger.hpp"
#include "base/exception.hpp"
#include "base/application.hpp"
#include "base/statsfunction.hpp"

using namespace icinga;

REGISTER_TYPE(HLMonReceiver);

REGISTER_STATSFUNCTION(HLMonReceiver, &HLMonReceiver::StatsFunc);

void HLMonReceiver::StatsFunc(const Dictionary::Ptr& status, const Array::Ptr&) {
	Dictionary::Ptr nodes = new Dictionary();

	for (const HLMonReceiver::Ptr& hlmonreceiver : ConfigType::GetObjectsByType<HLMonReceiver>()) {
		nodes->Set(hlmonreceiver->GetName(), 1); //add more stats
	}

	status->Set("hlmonreceiver", nodes);
}

void HLMonReceiver::Start(bool runtimeCreated) {
	ObjectImpl<HLMonReceiver>::Start(runtimeCreated);

	m_ReceiveThread = std::thread(std::bind(&HLMonReceiver::ReceiveThread, this));
	m_ReceiveThread.detach();
}

AmqpClient::Channel::ptr_t HLMonReceiver::getChannel() {
	AmqpClient::Channel::ptr_t channel;

	channel = AmqpClient::Channel::Create(GetHost(), GetPort(), GetUser(), GetPassword());
	AmqpClient::Table arguments;
	arguments.insert(AmqpClient::TableEntry(AmqpClient::TableKey("x-message-ttl"), AmqpClient::TableValue(900000)));
	channel->DeclareQueue(GetQueue(), false, true, false, false, arguments);
	channel->BasicConsume(GetQueue(), "", true, true, false, 0);

	return channel;
}

void HLMonReceiver::ReceiveThread() {
	Utility::SetThreadName("HL Monitoring Module Receiver");

	for (;;) {
		try {
			AmqpClient::Channel::ptr_t channel = HLMonReceiver::getChannel();

			for (;;) {
				try {
					AmqpClient::BasicMessage::ptr_t message = channel->BasicConsumeMessage()->Message();

					Log(LogDebug, "HLMonReceiver")
					<< "Message via RabbitMQ received: " << message->Body();

					try {
						String command = message->Body();
						Log(LogDebug, "HLMonReceiver")
						<< "Executing external command: " << command;

						ExternalCommandProcessor::Execute(command);
					} catch (const std::exception& ex) {
						Log(LogWarning, "HLMonReceiver")
						<< "External command failed: " << DiagnosticInformation(ex);
					}
				} catch (const std::exception& ex) {
					Log(LogCritical, "HLMonReceiver")
					<< "Unexpected error while receiving from RabbitMQ channel: " << ex.what();

					Utility::Sleep(1);
					break;
				}
			}
		} catch (const std::exception& ex) {
			Log(LogCritical, "HLMonReceiver")
			<< "Unexpected error while getting RabbitMQ channel: " << ex.what();

			Utility::Sleep(1);
		}
	}
}
