/******************************************************************************
 * Icinga 2                                                                   *
 * Copyright (C) 2018 Herrmann & Lenz Solutions GmbH (https://hl-solutions.de)*
 *                                                                            *
 * This program is free software; you can redistribute it and/or              *
 * modify it under the terms of the GNU General Public License                *
 * as published by the Free Software Foundation; either version 2             *
 * of the License, or (at your option) any later version.                     *
 *                                                                            *
 * This program is distributed in the hope that it will be useful,            *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of             *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              *
 * GNU General Public License for more details.                               *
 *                                                                            *
 * You should have received a copy of the GNU General Public License          *
 * along with this program; if not, write to the Free Software Foundation     *
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.             *
 ******************************************************************************/

#include "hlmon/hlmonsender.hpp"
#include "hlmon/hlmonsender-ti.cpp"

#include <SimpleAmqpClient/SimpleAmqpClient.h>

#include "icinga/service.hpp"
#include "icinga/checkcommand.hpp"
#include "icinga/eventcommand.hpp"
#include "icinga/notification.hpp"
#include "icinga/macroprocessor.hpp"
#include "icinga/externalcommandprocessor.hpp"
#include "icinga/compatutility.hpp"
#include "icinga/pluginutility.hpp"
#include "base/configtype.hpp"
#include "base/objectlock.hpp"
#include "base/logger.hpp"
#include "base/exception.hpp"
#include "base/convert.hpp"
#include "base/application.hpp"
#include "base/utility.hpp"
#include "base/statsfunction.hpp"

// JsonEncode, JsonDecode
#include "base/json.hpp"

using namespace icinga;

REGISTER_TYPE(HLMonSender);

REGISTER_STATSFUNCTION(HLMonSender, &HLMonSender::StatsFunc);

void HLMonSender::StatsFunc(const Dictionary::Ptr& status, const Array::Ptr&) {
	Dictionary::Ptr nodes = new Dictionary();

	for (const HLMonSender::Ptr& hlmonsender : ConfigType::GetObjectsByType<HLMonSender>()) {
		nodes->Set(hlmonsender->GetName(), 1); //add more stats
	}

	status->Set("hlmonsender", nodes);
}

/**
 * @threadsafety Always.
 */
void HLMonSender::Start(bool runtimeCreated) {
	ObjectImpl<HLMonSender>::Start(runtimeCreated);

	Checkable::OnNewCheckResult.connect(std::bind(&HLMonSender::CheckResultHandler, this, _1, _2));
	Checkable::OnAcknowledgementSet.connect(std::bind(&HLMonSender::SetAcknowledgementHandler, this, _1));
	Checkable::OnAcknowledgementCleared.connect(std::bind(&HLMonSender::ClearAcknowledgementHandler, this, _1));

	/*
	 * only for downtimes and acknowledgements, these operations have their own handlers
	 * comments are not used by the HL Monitoring Module
	Comment::OnCommentAdded.connect(std::bind(&HLMonSender::AddCommentHandler, this, _1));
	Comment::OnCommentRemoved.connect(std::bind(&HLMonSender::RemoveCommentHandler, this, _1));
	*/

	Downtime::OnDowntimeAdded.connect(std::bind(&HLMonSender::AddDowntimeHandler, this, _1));
	Downtime::OnDowntimeTriggered.connect(std::bind(&HLMonSender::TriggerDowntimeHandler, this, _1));
	Downtime::OnDowntimeRemoved.connect(std::bind(&HLMonSender::RemoveDowntimeHandler, this, _1));

	ConfigObject::OnVersionChanged.connect(std::bind(&HLMonSender::ChangeConfigHandler, this));
}

AmqpClient::Channel::ptr_t HLMonSender::getChannel() {
	AmqpClient::Channel::ptr_t channel;

	channel = AmqpClient::Channel::Create(GetHost(), GetPort(), GetUser(), GetPassword());
    AmqpClient::Table arguments;
    arguments.insert(AmqpClient::TableEntry(AmqpClient::TableKey("x-message-ttl"), AmqpClient::TableValue(900000)));
	channel->DeclareQueue(GetQueue(), false, true, false, false, arguments);
//	channel->BindQueue(GetQueue(), "");

	return channel;
}

/**
 * @threadsafety Always.
 */
void HLMonSender::CheckResultHandler(const Checkable::Ptr& checkable, const CheckResult::Ptr& cr) {
	HLMonSender::GenerateAndSendEvent(checkable, cr);
}

/**
 * @threadsafety Always.
 */
void HLMonSender::SetAcknowledgementHandler(const Checkable::Ptr& checkable) {
	HLMonSender::GenerateAndSendStat(checkable);
}

/**
 * @threadsafety Always.
 */
void HLMonSender::ClearAcknowledgementHandler(const Checkable::Ptr& checkable) {
	HLMonSender::GenerateAndSendStat(checkable);
}

/**
 * @threadsafety Always.
 */
void HLMonSender::AddCommentHandler(const Comment::Ptr& comment) {
	HLMonSender::GenerateAndSendStat(comment->GetCheckable());
}

/**
 * @threadsafety Always.
 */
void HLMonSender::RemoveCommentHandler(const Comment::Ptr& comment) {
	HLMonSender::GenerateAndSendStat(comment->GetCheckable());
}

/**
 * @threadsafety Always.
 */
void HLMonSender::AddDowntimeHandler(const Downtime::Ptr& downtime) {
	if (!downtime)
		return;

	HLMonSender::GenerateAndSendStat(downtime->GetCheckable());
}

/**
 * @threadsafety Always.
 */
void HLMonSender::TriggerDowntimeHandler(const Downtime::Ptr& downtime) {
	if (!downtime)
		return;

	HLMonSender::GenerateAndSendStat(downtime->GetCheckable());
}

/**
 * @threadsafety Always.
 */
void HLMonSender::RemoveDowntimeHandler(const Downtime::Ptr& downtime) {
	if (!downtime)
		return;

	HLMonSender::GenerateAndSendStat(downtime->GetCheckable());
}

/**
 * @threadsafety Always.
 */
void HLMonSender::ChangeConfigHandler(void) {
	HLMonSender::GenerateAndSendSchedulerStat();
}

/**
 * @threadsafety Always.
 */
void HLMonSender::GenerateAndSendEvent(const Checkable::Ptr& checkable, const CheckResult::Ptr& cr) {
	Host::Ptr host;
	Service::Ptr service;
	tie(host, service) = GetHostService(checkable);

	std::ostringstream msgbuf;

	if (service) {
		long event_timestamp = service->GetLastCheck();

		int customer_id = -1;
		int host_id = -1;
		int service_id = -1;
		Dictionary::Ptr serviceVars = service->GetVars();
		if (serviceVars) {
			ObjectLock olock (serviceVars);

			if (serviceVars->Get("customer_id"))
				customer_id = serviceVars->Get("customer_id");

			if (serviceVars->Get("host_id"))
				host_id = serviceVars->Get("host_id");

			if (serviceVars->Get("service_id"))
				service_id = serviceVars->Get("service_id");
		}

		String host_name = host->GetName();
		String service_name = service->GetShortName();
		int service_state_id = service->GetState();
		String service_state_type = Service::StateTypeToString(service->GetStateType());
		int service_attempt = service->GetCheckAttempt();
		int max_service_attempts = service->GetMaxCheckAttempts();

		CheckResult::Ptr check_result = service->GetLastCheckResult();
		double service_latency = 0, service_execution_time = 0;
		String service_output = "", service_long_output = "", service_perf_data = "";
		if (check_result) {
			service_latency = check_result->CalculateLatency();
			service_execution_time = check_result->CalculateExecutionTime();
			service_output = CompatUtility::GetCheckResultOutput(check_result);
            service_long_output = CompatUtility::GetCheckResultLongOutput(check_result);
            service_perf_data = PluginUtility::FormatPerfdata(cr->GetPerformanceData());
		}

		Dictionary::Ptr glob_srvc_evnt = new Dictionary();
		glob_srvc_evnt->Set("scheduler-id", Convert::ToString(GetScheduler()));
		glob_srvc_evnt->Set("EVENT_TIMESTAMP", Convert::ToString(event_timestamp));
		glob_srvc_evnt->Set("INTERNALCUSTID", Convert::ToString(customer_id));
		glob_srvc_evnt->Set("INTERNALHOSTID", Convert::ToString(host_id));
		glob_srvc_evnt->Set("INTERNALSRVCID", Convert::ToString(service_id));
		glob_srvc_evnt->Set("HOSTNAME", host_name);
		glob_srvc_evnt->Set("SERVICENAME", service_name);
		glob_srvc_evnt->Set("SERVICESTATEID", Convert::ToString(service_state_id));
		glob_srvc_evnt->Set("SERVICESTATETYPE", service_state_type);
		glob_srvc_evnt->Set("SERVICEATTEMPT", Convert::ToString(service_attempt));
		glob_srvc_evnt->Set("MAXSERVICEATTEMPTS", Convert::ToString(max_service_attempts));
		glob_srvc_evnt->Set("SERVICELATENCY", Convert::ToString(service_latency));
		glob_srvc_evnt->Set("SERVICEEXECUTIONTIME", Convert::ToString(service_execution_time));
		glob_srvc_evnt->Set("SERVICEOUTPUT", CompatUtility::UnEscapeString(service_output));
		glob_srvc_evnt->Set("LONGSERVICEOUTPUT", CompatUtility::UnEscapeString(service_long_output));
		glob_srvc_evnt->Set("SERVICEPERFDATA", CompatUtility::UnEscapeString(service_perf_data));

		Dictionary::Ptr json = new Dictionary();
		json->Set("GLOB_SRVC_EVNT", glob_srvc_evnt);

		msgbuf << JsonEncode(json);
	}
	else {
		long event_timestamp = host->GetLastCheck();

		int customer_id = -1;
		int host_id = -1;
		Dictionary::Ptr hostVars = host->GetVars();
		if (hostVars) {
			ObjectLock olock (hostVars);

			if (hostVars->Get("customer_id"))
				customer_id = hostVars->Get("customer_id");

			if (hostVars->Get("host_id"))
				host_id = hostVars->Get("host_id");
		}

		String host_name = host->GetName();
		int host_state_id = host->GetState();
		String host_state_type = Host::StateTypeToString(host->GetStateType());
		int host_attempt = host->GetCheckAttempt();
		int max_host_attempts = host->GetMaxCheckAttempts();

		CheckResult::Ptr check_result = host->GetLastCheckResult();
		double host_latency = 0, host_execution_time = 0;
		String host_output = "", host_long_output = "", host_perf_data = "";
		if (check_result) {
			host_latency = check_result->CalculateLatency();
			host_execution_time = check_result->CalculateExecutionTime();
			host_output = CompatUtility::GetCheckResultOutput(check_result);
			host_long_output = CompatUtility::GetCheckResultLongOutput(check_result);
			host_perf_data = PluginUtility::FormatPerfdata(cr->GetPerformanceData());
		}

		Dictionary::Ptr glob_host_evnt = new Dictionary();
		glob_host_evnt->Set("scheduler-id", Convert::ToString(GetScheduler()));
		glob_host_evnt->Set("EVENT_TIMESTAMP", Convert::ToString(event_timestamp));
		glob_host_evnt->Set("INTERNALCUSTID", Convert::ToString(customer_id));
		glob_host_evnt->Set("INTERNALHOSTID", Convert::ToString(host_id));
		glob_host_evnt->Set("HOSTNAME", host_name);
		glob_host_evnt->Set("HOSTSTATEID", Convert::ToString(host_state_id));
		glob_host_evnt->Set("HOSTSTATETYPE", host_state_type);
		glob_host_evnt->Set("HOSTATTEMPT", Convert::ToString(host_attempt));
		glob_host_evnt->Set("MAXHOSTATTEMPTS", Convert::ToString(max_host_attempts));
		glob_host_evnt->Set("HOSTLATENCY", Convert::ToString(host_latency));
		glob_host_evnt->Set("HOSTEXECUTIONTIME", Convert::ToString(host_execution_time));
		glob_host_evnt->Set("HOSTOUTPUT", CompatUtility::UnEscapeString(host_output));
		glob_host_evnt->Set("LONGHOSTOUTPUT", CompatUtility::UnEscapeString(host_long_output));
		glob_host_evnt->Set("HOSTPERFDATA", CompatUtility::UnEscapeString(host_perf_data));

		Dictionary::Ptr json = new Dictionary();
		json->Set("GLOB_HOST_EVNT", glob_host_evnt);

		msgbuf << JsonEncode(json);
	}

	String jsonMessage = msgbuf.str();

	try {
		ObjectLock olock(this);
		SendToRabbitMQ(jsonMessage);
	} catch (const std::exception& ex) {
		Log(LogCritical, "HLMonSender")
		<< "Unexpected error while sending to RabbitMQ channel: " << ex.what();

		Utility::Sleep(1);

		try {
			channel = getChannel();
		} catch (const std::exception& ex) {}
	}
}

/**
 * @threadsafety Always.
 */
void HLMonSender::GenerateAndSendStat(const Checkable::Ptr& checkable) {
	Host::Ptr host;
	Service::Ptr service;
	tie(host, service) = GetHostService(checkable);

	std::ostringstream msgbuf;

	if (service) {
		Dictionary::Ptr glob_srvc_stat = new Dictionary();
		glob_srvc_stat->Set("scheduler-id", Convert::ToString(GetScheduler()));
		glob_srvc_stat->Set("SERVICENAME", Convert::ToString(service->GetShortName()));
		glob_srvc_stat->Set("INTERNALSRVCID", Convert::ToString(service->GetVars()->Get("service_id")));
		glob_srvc_stat->Set("ACKNOWLEDGED", service->GetAcknowledgement() != AcknowledgementNone ? "Y" : "N");

		std::set<Downtime::Ptr> downtimes = service->GetDowntimes();
		for (const Downtime::Ptr& downtime : downtimes) {
			if (downtime->IsExpired())
				continue;

			Dictionary::Ptr glob_srvc_stat_dwnt = new Dictionary();
			glob_srvc_stat_dwnt->Set("downtime-id", Convert::ToString(downtime->GetLegacyId()));
			glob_srvc_stat_dwnt->Set("downtime-name", Convert::ToString(downtime->GetName()));
			glob_srvc_stat_dwnt->Set("ENTRYTIME", Convert::ToString((long) downtime->GetEntryTime()));
			glob_srvc_stat_dwnt->Set("STARTTIME", Convert::ToString((long) downtime->GetStartTime()));
			glob_srvc_stat_dwnt->Set("ENDTIME", Convert::ToString((long) downtime->GetEndTime()));
			glob_srvc_stat_dwnt->Set("FIXED", downtime->GetFixed() ? "Y" : "N");
			glob_srvc_stat_dwnt->Set("DURATION", Convert::ToString((long) downtime->GetDuration()));
			glob_srvc_stat_dwnt->Set("AUTHOR", Convert::ToString(downtime->GetAuthor()));
			glob_srvc_stat_dwnt->Set("COMMENT", Convert::ToString(downtime->GetComment()));

			Array::Ptr list = glob_srvc_stat->Get("GLOB_SRVC_STAT_DWNT");
			if (!list) {
				list = new Array();
				glob_srvc_stat->Set("GLOB_SRVC_STAT_DWNT", list);
			}
			list->Add(glob_srvc_stat_dwnt);
		}

		std::set<Comment::Ptr> comments = service->GetComments();
		for (const Comment::Ptr& comment : comments) {
			if (comment->IsExpired())
				continue;

			Dictionary::Ptr glob_srvc_stat_cmnt = new Dictionary();
			glob_srvc_stat_cmnt->Set("comment-id", Convert::ToString(comment->GetLegacyId()));
			glob_srvc_stat_cmnt->Set("comment-name", Convert::ToString(comment->GetName()));

			Array::Ptr list = glob_srvc_stat->Get("GLOB_SRVC_STAT_CMNT");
			if (!list) {
				list = new Array();
				glob_srvc_stat->Set("GLOB_SRVC_STAT_CMNT", list);
			}
			list->Add(glob_srvc_stat_cmnt);
		}

		Dictionary::Ptr json = new Dictionary();
		json->Set("GLOB_SRVC_STAT", glob_srvc_stat);
		msgbuf << JsonEncode(json);
		String jsonMessage = msgbuf.str();
	}
	else {
		Dictionary::Ptr glob_host_stat = new Dictionary();
		glob_host_stat->Set("scheduler-id", Convert::ToString(GetScheduler()));
		glob_host_stat->Set("HOSTNAME", Convert::ToString(host->GetName()));
		glob_host_stat->Set("INTERNALHOSTID", Convert::ToString(host->GetVars()->Get("host_id")));
		glob_host_stat->Set("ACKNOWLEDGED", host->GetAcknowledgement() != AcknowledgementNone ? "Y" : "N");

		std::set<Downtime::Ptr> downtimes = host->GetDowntimes();
		for (const Downtime::Ptr& downtime : downtimes) {
			if (downtime->IsExpired())
				continue;

			Dictionary::Ptr glob_host_stat_dwnt = new Dictionary();
			glob_host_stat_dwnt->Set("downtime-id", Convert::ToString(downtime->GetLegacyId()));
			glob_host_stat_dwnt->Set("downtime-name", Convert::ToString(downtime->GetName()));
			glob_host_stat_dwnt->Set("ENTRYTIME", Convert::ToString((long) downtime->GetEntryTime()));
			glob_host_stat_dwnt->Set("STARTTIME", Convert::ToString((long) downtime->GetStartTime()));
			glob_host_stat_dwnt->Set("ENDTIME", Convert::ToString((long) downtime->GetEndTime()));
			glob_host_stat_dwnt->Set("FIXED", downtime->GetFixed() ? "Y" : "N");
			glob_host_stat_dwnt->Set("DURATION", Convert::ToString((long) downtime->GetDuration()));
			glob_host_stat_dwnt->Set("AUTHOR", Convert::ToString(downtime->GetAuthor()));
			glob_host_stat_dwnt->Set("COMMENT", Convert::ToString(downtime->GetComment()));

			Array::Ptr list = glob_host_stat->Get("GLOB_HOST_STAT_DWNT");
			if (!list) {
				list = new Array();
				glob_host_stat->Set("GLOB_HOST_STAT_DWNT", list);
			}
			list->Add(glob_host_stat_dwnt);
		}

		std::set<Comment::Ptr> comments = host->GetComments();
		for (const Comment::Ptr& comment : comments) {
			Dictionary::Ptr glob_host_stat_cmnt = new Dictionary();
			glob_host_stat_cmnt->Set("comment-id", Convert::ToString(comment->GetLegacyId()));
            glob_host_stat_cmnt->Set("comment-name", Convert::ToString(comment->GetName()));

			Array::Ptr list = glob_host_stat->Get("GLOB_HOST_STAT_CMNT");
			if (!list) {
				list = new Array();
				glob_host_stat->Set("GLOB_HOST_STAT_CMNT", list);
			}
			list->Add(glob_host_stat_cmnt);
		}

		Dictionary::Ptr json = new Dictionary();
		json->Set("GLOB_HOST_STAT", glob_host_stat);
		msgbuf << JsonEncode(json);
	}

	String jsonMessage = msgbuf.str();

	try {
		ObjectLock olock(this);
		SendToRabbitMQ(jsonMessage);
	} catch (const std::exception& ex) {
		Log(LogCritical, "HLMonSender")
		<< "Unexpected error while sending to RabbitMQ channel: " << ex.what();

		Utility::Sleep(1);

		try {
			channel = getChannel();
		} catch (const std::exception& ex) {}
	}
}

/**
 * @threadsafety Always.
 */
void HLMonSender::GenerateAndSendSchedulerStat(void) {
	std::ostringstream msgbuf;

	Dictionary::Ptr glob_schd_stat = new Dictionary();
	glob_schd_stat->Set("scheduler-id", Convert::ToString(GetScheduler()));
	glob_schd_stat->Set("NOTIFICATIONS", IcingaApplication::GetInstance()->GetEnableNotifications() ? "Y" : "N");

	Dictionary::Ptr json = new Dictionary();
	json->Set("GLOB_SCHD_STAT", glob_schd_stat);
	msgbuf << JsonEncode(json);
	String jsonMessage = msgbuf.str();

	try {
		ObjectLock olock(this);
		SendToRabbitMQ(jsonMessage);
	} catch (const std::exception& ex) {
		Log(LogCritical, "HLMonSender")
		<< "Unexpected error while sending to RabbitMQ channel: " << ex.what();

		Utility::Sleep(1);

		try {
			channel = getChannel();
		} catch (const std::exception& ex) {}
	}
}

void HLMonSender::SendToRabbitMQ(const String& jsonMessage) {
	ASSERT(OwnsLock());

	if (!channel)
		channel = getChannel();

	AmqpClient::BasicMessage::ptr_t message = AmqpClient::BasicMessage::Create();
	message->Body(jsonMessage);

	Log(LogDebug, "HLMonSender")
	<< "Sending JSON to RabbitMQ channel: " << jsonMessage;

	channel->BasicPublish("", GetQueue(), message);
}
